#pragma once

#include "string_view.h"
#include "Owned_string.h"

#include <cstddef>

class Shared_string
{
public:
    Shared_string();

    Shared_string(Shared_string const&);
    Shared_string(Shared_string&&) noexcept;
    Shared_string(string_view);
    Shared_string(Owned_string&&);

    ~Shared_string();

    Shared_string& operator=(Shared_string const&);
    Shared_string& operator=(Shared_string&&) noexcept;
    Shared_string& operator=(string_view);
    Shared_string& operator=(Owned_string&&);

    char operator[](size_t) const;

    Shared_string substr(size_t start) const;
    Shared_string substr(size_t start, size_t size) const;

    size_t size() const;
    bool empty() const;

    char const* begin() const;
    char const* end() const;

    char const* c_str() const;

    operator string_view() const;
    operator Owned_string() &&;

    size_t ref_count() const;

private:
    void decrement_() const;
    void increment_() const;
    void pod_assign_(Shared_string const&);

    // This order of member variables is essential for correct
    // initialization on the Shared_string(Owned_string&&)
    // constructor, which is actually the only way that we allocate
    // for shared strings. This order ensures that our first
    // initialization step is to allocate the ref count, which
    // is the only step that can fail, and thus why we do it first.
    // If it fails then the `Owned_string` argument still owns
    // the memory. Next we initialize view_ based on the range
    // that the Owned_string is about to forget when we steal its
    // memory with `Owned_string::release()`.
    size_t*     ref_count_;
    string_view view_;
    char*       data_;
};
