#include "Uniq_ptr.h"
#include "tracker.h"
#include <catch.h>

TEST_CASE("null pointer")
{
    Uniq_ptr<int> np;
    CHECK_FALSE( np );
    CHECK( np == nullptr );
}

TEST_CASE("Construct from new pointer")
{
    Uniq_ptr<int> up(new int(5));
    CHECK( *up == 5 );
    CHECK( up );
    CHECK( up != nullptr );
}

TEST_CASE("Construct from make_Uniq")
{
    Uniq_ptr<int> up = make_Uniq<int>(7);
    CHECK( *up == 7 );
    CHECK( up );
    CHECK( up != nullptr );
}

TEST_CASE("Moving")
{
    Uniq_ptr<int> a = make_Uniq<int>(8);
    CHECK( a );
    Uniq_ptr<int> b = std::move(a);
    CHECK_FALSE( a );
    CHECK( b );
    CHECK( *b == 8 );

    a.reset(new int(17));
    CHECK( *a == 17 );
    CHECK( *b == 8 );

    int* raw_a = &*a;
    int* raw_b = &*b;

    using std::swap;

    swap(a, b);
    CHECK( *a == 8 );
    CHECK( *b == 17 );
    CHECK( &*a == raw_b );
    CHECK( &*b == raw_a );

    swap(*a, *b);
    CHECK( *a == 17 );
    CHECK( *b == 8 );
    CHECK( &*a == raw_b );
    CHECK( &*b == raw_a );

    b = std::move(a);
    CHECK( *b == 17 );
    CHECK( &*b == raw_b );
    CHECK( a == nullptr );
}

TEST_CASE("struct")
{
    using Tracker = Tracker<int>;
    using ptr_t   = Tracker::ptr_t;

    Tracker track;

    auto expect = [&](Tracker::log_t expected) {
        CHECK( track.log() == expected );
    };

    track(3);
    expect({3});

    {
        ptr_t p1 = track(5);
        expect({3});

        ptr_t p2 = track(7);
        expect({3});

        p2.reset();
        expect({3, 7});

        {
            p2 = track(9);
            ptr_t p3 = track(11);
            expect({3, 7});
        }

        expect({3, 7, 11});
    }

    expect({3, 7, 11, 9, 5});
}

