#include "fn_tracker.h"
#include <catch.h>
#include <memory>
#include <vector>

using log_t   = std::vector<int>;
using track_t = Tracker<std::shared_ptr, log_t>;
using ptr_t   = track_t::ptr_t;

TEST_CASE("Tracker")
{
    track_t track;

    auto expect = [&](log_t expected) {
        CHECK( track.get_state() == expected );
    };

    track << 3;
    expect({3});

    {
        ptr_t p1 = track << 5;
        expect({3});

        ptr_t p2 = track << 7;
        expect({3});

        p2.reset();
        expect({3, 7});

        {
            p2 = track << 9;
            ptr_t p3 = track << 11;
            expect({3, 7});
        }

        expect({3, 7, 11});

        CHECK( p1->get_state().size() == 3 );
    }

    expect({3, 7, 11, 9, 5});
}

TEST_CASE("Outlive")
{
    ptr_t p1, p2, p3;

    {
        track_t track;

        p1 = track << 9;
        p2 = track << 8;
        p3 = track << 7;

        CHECK( track.get_state().empty() );

        p1.reset();
        CHECK( track.get_state().size() == 1 );
    }


    CHECK( p3->get_state().size() == 1 );
    p2 = {};

    auto& r3 = *p3;
    CHECK( r3.get_state().size() == 2 );

    {
        ptr_t p4 = r3 << 6;
        CHECK( r3.get_state().size() == 2 );
        p3 = std::move(p4);
        CHECK( p3->get_state()[0] );
    }

    ptr_t p5 = *p3 << 5;

    p3 = nullptr;
    CHECK( p5->get_state() == log_t{9, 8, 7, 6} );
}

